import requests
from json.decoder import JSONDecodeError

from ..errors import APIError, PermissionsError


def _parse_resp(resp):
    if resp.status_code == 403:
        raise PermissionsError(resp.json())
    if resp.status_code != 200:
        raise APIError(resp.text)
    try:
        data = resp.json()
        if data["s"] != True:
            raise APIError(data["m"])
        return data
    except JSONDecodeError:
        return resp.content

def get(api_path, ctf, *args, **kwargs):
    if ctf.auth_token:
        kwargs["headers"] = {}
        kwargs["headers"]["authorization"] = ctf.auth_token
    resp = requests.get((ctf.api_base if not "http" in api_path else "") + api_path, *args, **kwargs)
    return _parse_resp(resp)


def post(api_path, ctf, *args, **kwargs):
    if ctf.auth_token:
        kwargs["headers"] = {}
        kwargs["headers"]["authorization"] = ctf.auth_token
    resp = requests.post((ctf.api_base if not "http" in api_path else "") + api_path, *args, **kwargs)
    return _parse_resp(resp)


def patch(api_path, values, ctf, *args, **kwargs):
    if ctf.auth_token:
        kwargs["headers"] = {}
        kwargs["headers"]["authorization"] = ctf.auth_token
    resp = requests.patch((ctf.api_base if not "http" in api_path else "") + api_path, values, *args, **kwargs)
    return _parse_resp(resp)



import click

from .logic import clone, submit

@click.group()
def cli():
    pass

@cli.command()
@click.argument('api_url', required=True)
def clone(api_url):
    clone(api_url)

@cli.command()
@click.argument('flag', required=True)
def submit(flag):
    submit(flag)


import os
import json
from shutil import rmtree

from requests import exceptions
import getpass

from .utils import log
from .utils import files
from .ractf import CTF

DO_NOT_TOUCH_COMMENT = "This file is automatically generated and used by RACLI to track state. Don't modify it unless you know what you're doing."

def init_ctf():
    ctf_info = files.find_file(".ctf.json")
    if not ctf_info:
        log.error_and_quit("Could not find .ctf.json, have you cloned?")

    with open(ctf_info, "r") as f:
        ctf_info = json.load(f)

    ctf = CTF(ctf_info["api_url"])
    ctf.auth_token = f"{ctf_info['token']}"

    return ctf

def clone(api_url):
    ctf = CTF(api_url)

    ctf.get_config()

    try:
        clone_into = ctf.get_config()["flag_prefix"]
    except (exceptions.ConnectionError, ractf.errors.APIError, KeyError):
        log.error_and_quit("Failed to get CTF information from that URL. Please ensure you have supplied a valid API base.")

    if os.path.exists(clone_into):
        log.error_and_quit(f"Path '{clone_into}' already exists. Aborting.")

    log.info(f"Cloning CTF into '{clone_into}'...")

    username = input("Username: ")
    password = getpass.getpass("Password: ")
    otp = getpass.getpass("OTP (Leave blank if none): ")

    try:
        ctf.login(username=username,password=password,otp=otp)
    except ractf.errors.APIError:
        log.error_and_quit("Could not login with supplied credentials")
    
    ctf_store = {
        "_comment": DO_NOT_TOUCH_COMMENT,
        "token": ctf.auth_token,
        "api_url": ctf.api_base,
    }
    
    os.mkdir(clone_into)
    
    for category in ctf.get_categories():
        os.mkdir(f"{clone_into}/{category.name}")

        with open(f"{clone_into}/{category.name}/.category.json", "w+") as f:
            json.dump({"_comment": DO_NOT_TOUCH_COMMENT, "id": category.id}, f)

        for challenge in category.challenges:
            if challenge.unlocked:
                os.mkdir(f"{clone_into}/{category.name}/{challenge.name}")

                with open(f"{clone_into}/{category.name}/{challenge.name}/brief.md", "w+", encoding="utf-8") as f:
                    f.write(f"# {challenge.name}\n\n{challenge.description}")

                with open(f"{clone_into}/{category.name}/{challenge.name}/.challenge.json", "w+") as f:
                    json.dump({"_comment": DO_NOT_TOUCH_COMMENT, "id": challenge.id}, f)
            else:
                os.mkdir(f"{clone_into}/{category.name}/Locked challenge [ID {challenge.id}]")

    with open(f"{clone_into}/.ctf.json", "w+") as f:
        json.dump(ctf_store, f)
    
    log.success("CTF Cloned!")
    
def submit(flag):
    ctf = init_ctf()

    challenge_info = files.find_file(".challenge.json")
    if not challenge_info:
        log.error_and_quit("Could not find .challenge.json, are you in a challenge directory?")
    with open(challenge_info, "r") as f:
        challenge_info = json.load(f)

    challenge = ctf.get_challenge(challenge_info["id"])
    if challenge.submit_flag(flag):
        log.success("Correct flag submitted. Well done!")
        if len(challenge.unlocks) > 0:
            log.warn("You may have unlocked additional challenges, make sure to check.")
            for challenge in challenge.unlocks:
                challenge = ctf.get_challenge(challenge)
                cat_root = files.get_cat_root()
                rmtree(f"{cat_root}/Locked challenge [ID {challenge.id}]")

                os.mkdir(f"{cat_root}/{challenge.name}")

                with open(f"{cat_root}/{challenge.name}/brief.md", "w+", encoding="utf-8") as f:
                    f.write(f"# {challenge.name}\n\n{challenge.description}")

                with open(f"{cat_root}/{challenge.name}/.challenge.json", "w+") as f:
                    json.dump({"_comment": DO_NOT_TOUCH_COMMENT, "id": challenge.id}, f)
    else:
        log.info("Flag rejected. Keep trying!")



from .abc import APIBaseObject
from .helpers.requests import post
from .hint import Hint
from .file import File


class Challenge(APIBaseObject):
    def __init__(self, object_id, ctf, data=None):
        self.id = object_id
        self.ctf = ctf
        super(Challenge, self).__init__(ctf, data=data)
        self.convert_json_to_hints()
        self.convert_json_to_files()

    def get_api_path(self):
        return f"challenges/{self.id}/"

    def convert_json_to_hints(self):
        hints = []
        for hint in self.hints:
            if isinstance(hint, Hint):
                continue
            hints.append(Hint(hint["id"], self, self.ctf, data=hint))
        self.hints = hints
    
    def convert_json_to_files(self):
        files = []
        for instance in self.files:
            if isinstance(instance, File):
                continue
            files.append(File(instance["id"], self, self.ctf, data=instance))
        self.files = files
    
    def submit_flag(self, flag):
        resp = post("challenges/submit_flag/", self.ctf, json={"challenge": self.id, "flag": flag})
        # again, bug
        # self._fill_attrs()
        return resp["m"] == "correct_flag"


from .challenge import Challenge
from .category import Category
from .helpers.requests import get, post

class CTF:
    def __init__(self, api_base):
        self.api_base = api_base
        self.auth_token = None

    def get_challenge(self, id):
        # There's currently a bug in RACTF that means we can't GET /challenges/:id, so this is a workaround
        return [challenge for challenge in self.get_challenges() if challenge.id == id][0]
        # return Challenge(id, self)

    def get_challenges(self):
        resp = get("challenges/categories/", self)
        challenges = []
        for category in resp["d"]:
            for challenge in category["challenges"]:
                challenges.append(Challenge(challenge["id"], self, data=challenge))
        return challenges
    
    def get_categories(self):
        resp = get("challenges/categories/", self)
        categories = []
        for category in resp["d"]:
            categories.append(Category(category["id"], self, data=category))
        return categories

    def get_config(self):
        return get("config/", self)["d"]
    
    def login(self, username, password, otp=None):
        self.auth_token = "Token " + post("auth/login/", self, json={"username": username, "password": password, "otp": otp})["d"]["token"]


class bcolors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'

def info(text):
    print(text)
def warn(text):
    print(bcolors.WARNING+text+bcolors.ENDC)
def success(text):
    print(bcolors.OKGREEN+text+bcolors.ENDC)
def error_and_quit(text):
    print(bcolors.FAIL + text + bcolors.ENDC)
    raise SystemExit(1)


from .abc import APIBaseObject
from .helpers.requests import post

class Hint(APIBaseObject):
    def __init__(self, object_id, challenge, ctf, data=None):
        self.id = object_id
        self.ctf = ctf
        self.challenge = challenge
        self.text = None
        super(Hint, self).__init__(ctf, data=data)

    def get_api_path(self):
        return f"hints/{self.id}/"

    def use(self):
        self._fill_attrs(data=post("hints/use/", self.ctf, json={"id": self.id})["d"])



from .abc import APIBaseObject
from .challenge import Challenge


class Category(APIBaseObject):
    def __init__(self, object_id, ctf, data=None):
        self.id = object_id
        self.ctf = ctf
        super(Category, self).__init__(ctf, data=data)
        self.convert_json_to_challenges()

    def get_api_path(self):
        return f"challenges/categories/{self.id}/"
    
    def convert_json_to_challenges(self):
        challenges = []
        for challenge in self.challenges:
            if isinstance(challenge, Challenge):
                continue
            challenges.append(Challenge(challenge["id"], self.ctf, challenge))
        self.challenges = challenges


import os
from pathlib import Path

def find_file(filename, path=os.getcwd()):
    if os.path.dirname(path) == path:
        return None
    if os.path.exists(path + "/" + filename):
        return path + "/" + filename
    return find_file(filename, path=str(Path(path).parent))

def get_ctf_root():
    return find_file(".ctf.json").strip("/.ctf.json")

def get_cat_root():
    return find_file(".category.json").strip("/.category.json")


from .abc import APIBaseObject
from .helpers.requests import get

class File(APIBaseObject):
    def __init__(self, object_id, challenge, ctf, data=None):
        self.id = object_id
        self.ctf = ctf
        self.challenge = challenge
        super(File, self).__init__(ctf, data=data)

    def get_api_path(self):
        return f"files/{self.id}/"

    def save(self, path):
        with open(path, "wb") as f:
            f.write(self.read())

    def read(self):
        return get(self.url, self.ctf)


from abc import ABC, abstractmethod

from .helpers.requests import get, patch


class APIBaseObject(ABC):
    def __init__(self, ctf, data=None):
        self._ctf = ctf
        self._type = self.__class__.__name__
        self._fill_attrs(data=data)

    def __repr__(self):
        return f"<{self._type} [{self.id}]>"
    
    def __str__(self):
        return self.name

    @abstractmethod
    def get_api_path(self):
        raise NotImplementedError()

    def _set_attrs_on_remote(self, values):
        patch(self.get_api_path(), values, self._ctf)

    def _set_attr_on_remote(self, name, value):
        self._set_attrs({name: value})

    def _get_object(self):
        return get(self.get_api_path(), self._ctf)

    def _fill_attrs(self, data=None):
        if not data:
            data = self._get_object()
        for k, v in data.items():
            setattr(self, k, v)
